@@ -7,23 +7,28 @@
 # License:  Standard 3-clause BSD; see "license.txt" for full license terms
 #           and contributor agreement.
 
-from __future__ import absolute_import
+"""
+    Implements support for high-level access to HDF5 groups.
+"""
+
+from __future__ import absolute_import, division
 
 import posixpath as pp
-
 import six
-
 import numpy
-import collections
 
-from .. import h5g, h5i, h5o, h5r, h5t, h5l, h5p
+
+from .compat import filename_decode, filename_encode
+
+from .. import h5, h5g, h5i, h5o, h5r, h5t, h5l, h5p, h5s, h5d
 from . import base
-from .base import HLObject, MutableMappingWithLock, phil, with_phil
+from .base import HLObject, MutableMappingHDF5, phil, with_phil
 from . import dataset
 from . import datatype
+from .vds import vds_support
 
 
-class Group(HLObject, MutableMappingWithLock):
+class Group(HLObject, MutableMappingHDF5):
 
     """ Represents an HDF5 group.
     """
@@ -36,15 +41,31 @@ class Group(HLObject, MutableMappingWithLock):
                 raise ValueError("%s is not a GroupID" % bind)
             HLObject.__init__(self, bind)
 
-    def create_group(self, name):
+
+    _gcpl_crt_order = h5p.create(h5p.GROUP_CREATE)
+    _gcpl_crt_order.set_link_creation_order(
+        h5p.CRT_ORDER_TRACKED | h5p.CRT_ORDER_INDEXED)
+    _gcpl_crt_order.set_attr_creation_order(
+        h5p.CRT_ORDER_TRACKED | h5p.CRT_ORDER_INDEXED)
+
+
+    def create_group(self, name, track_order=None):
         """ Create and return a new subgroup.
 
         Name may be absolute or relative.  Fails if the target name already
         exists.
+
+        track_order
+            Track dataset/group/attribute creation order under this group
+            if True. If None use global default h5.get_config().track_order.
         """
+        if track_order is None:
+            track_order = h5.get_config().track_order
+
         with phil:
             name, lcpl = self._e(name, lcpl=True)
-            gid = h5g.create(self.id, name, lcpl=lcpl)
+            gcpl = Group._gcpl_crt_order if track_order else None
+            gid = h5g.create(self.id, name, lcpl=lcpl, gcpl=gcpl)
             return Group(gid)
 
     def create_dataset(self, name, shape=None, dtype=None, data=None, **kwds):
@@ -98,7 +119,19 @@ class Group(HLObject, MutableMappingWithLock):
             (Scalar) Use this value for uninitialized parts of the dataset.
         track_times
             (T/F) Enable dataset creation timestamps.
+        track_order
+            (T/F) Track attribute creation order if True. If omitted use
+            global default h5.get_config().track_order.
+        external
+            (List of tuples) Sets the external storage property, thus
+            designating that the dataset will be stored in one or more
+            non-HDF5 file(s) external to the HDF5 file. Adds each listed
+            tuple of (file[, offset[, size]]) to the dataset's list of
+            external files.
         """
+        if 'track_order' not in kwds:
+            kwds['track_order'] = h5.get_config().track_order
+
         with phil:
             dsid = dataset.make_new_dset(self, shape, dtype, data, **kwds)
             dset = dataset.Dataset(dsid)
@@ -106,6 +139,38 @@ class Group(HLObject, MutableMappingWithLock):
                 self[name] = dset
             return dset
 
+    if vds_support:
+        def create_virtual_dataset(self, name, layout, fillvalue=None):
+            """Create a new virtual dataset in this group.
+
+            Creates the virtual dataset from a list of virtual maps, any
+            gaps are filled with a specified fill value.
+
+            name
+                (str) Name of the new dataset
+
+            virtual_target
+                Defines the sources for the virtual dataset
+
+            """
+            from .vds import VDSmap
+            # Encode filenames and dataset names appropriately.
+            sources = [VDSmap(vspace, filename_encode(file_name),
+                              self._e(dset_name), src_space)
+                       for (vspace, file_name, dset_name, src_space)
+                       in layout.sources]
+
+            with phil:
+                dsid = dataset.make_new_virtual_dset(self, layout.shape,
+                         sources=sources, dtype=layout.dtype,
+                         maxshape=layout.maxshape, fillvalue=fillvalue)
+
+                dset = dataset.Dataset(dsid)
+                if name is not None:
+                    self[name] = dset
+
+            return dset
+
     def require_dataset(self, name, shape, dtype, exact=False, **kwds):
         """ Open a dataset, creating it if it doesn't exist.
 
@@ -138,8 +203,41 @@ class Group(HLObject, MutableMappingWithLock):
 
             return dset
 
+    def create_dataset_like(self, name, other, **kwupdate):
+        """ Create a dataset similar to `other`.
+
+        name
+            Name of the dataset (absolute or relative).  Provide None to make
+            an anonymous dataset.
+        other
+            The dataset which the new dataset should mimic. All properties, such
+            as shape, dtype, chunking, ... will be taken from it, but no data
+            or attributes are being copied.
+
+        Any dataset keywords (see create_dataset) may be provided, including
+        shape and dtype, in which case the provided values take precedence over
+        those from `other`.
+        """
+        for k in ('shape', 'dtype', 'chunks', 'compression',
+                  'compression_opts', 'scaleoffset', 'shuffle', 'fletcher32',
+                  'fillvalue'):
+            kwupdate.setdefault(k, getattr(other, k))
+        # TODO: more elegant way to pass these (dcpl to create_dataset?)
+        dcpl = other.id.get_create_plist()
+        kwupdate.setdefault('track_times', dcpl.get_obj_track_times())
+        kwupdate.setdefault('track_order', dcpl.get_attr_creation_order() > 0)
+
+        # Special case: the maxshape property always exists, but if we pass it
+        # to create_dataset, the new dataset will automatically get chunked
+        # layout. So we copy it only if it is different from shape.
+        if other.maxshape != other.shape:
+            kwupdate.setdefault('maxshape', other.maxshape)
+
+        return self.create_dataset(name, **kwupdate)
+
     def require_group(self, name):
-        """ Return a group, creating it if it doesn't exist.
+        # TODO: support kwargs like require_dataset
+        """Return a group, creating it if it doesn't exist.
 
         TypeError is raised if something with that name already exists that
         isn't a group.
@@ -195,8 +293,9 @@ class Group(HLObject, MutableMappingWithLock):
 
         >>> cls = group.get('foo', getclass=True)
         >>> if cls == SoftLink:
-        ...     print '"foo" is a soft link!'
         """
+        # pylint: disable=arguments-differ
+
         with phil:
             if not (getclass or getlink):
                 try:
@@ -225,19 +324,21 @@ class Group(HLObject, MutableMappingWithLock):
                         return SoftLink
                     linkbytes = self.id.links.get_val(self._e(name))
                     return SoftLink(self._d(linkbytes))
+
                 elif typecode == h5l.TYPE_EXTERNAL:
                     if getclass:
                         return ExternalLink
                     filebytes, linkbytes = self.id.links.get_val(self._e(name))
-                    # TODO: I think this is wrong,
-                    # we should use filesystem decoding on the filename
-                    return ExternalLink(self._d(filebytes), self._d(linkbytes))
+                    return ExternalLink(
+                        filename_decode(filebytes), self._d(linkbytes)
+                    )
+
                 elif typecode == h5l.TYPE_HARD:
                     return HardLink if getclass else HardLink()
+
                 else:
                     raise TypeError("Unknown link type")
 
-    @with_phil
     def __setitem__(self, name, obj):
         """ Add an object to the group.  The name must not already be in use.
 
@@ -262,26 +363,33 @@ class Group(HLObject, MutableMappingWithLock):
             values are stored as scalar datasets. Raise ValueError if we
             can't understand the resulting array dtype.
         """
-        name, lcpl = self._e(name, lcpl=True)
+        do_link = False
+        with phil:
+            name, lcpl = self._e(name, lcpl=True)
 
-        if isinstance(obj, HLObject):
-            h5o.link(obj.id, self.id, name, lcpl=lcpl, lapl=self._lapl)
+            if isinstance(obj, HLObject):
+                h5o.link(obj.id, self.id, name, lcpl=lcpl, lapl=self._lapl)
 
-        elif isinstance(obj, SoftLink):
-            self.id.links.create_soft(name, self._e(obj.path),
-                          lcpl=lcpl, lapl=self._lapl)
+            elif isinstance(obj, SoftLink):
+                self.id.links.create_soft(name, self._e(obj.path),
+                              lcpl=lcpl, lapl=self._lapl)
 
-        elif isinstance(obj, ExternalLink):
-            self.id.links.create_external(name, self._e(obj.filename),
-                          self._e(obj.path), lcpl=lcpl, lapl=self._lapl)
+            elif isinstance(obj, ExternalLink):
+                do_link = True
 
-        elif isinstance(obj, numpy.dtype):
-            htype = h5t.py_create(obj)
-            htype.commit(self.id, name, lcpl=lcpl)
+            elif isinstance(obj, numpy.dtype):
+                htype = h5t.py_create(obj, logical=True)
+                htype.commit(self.id, name, lcpl=lcpl)
 
-        else:
-            ds = self.create_dataset(None, data=obj, dtype=base.guess_dtype(obj))
-            h5o.link(ds.id, self.id, name, lcpl=lcpl)
+            else:
+                ds = self.create_dataset(None, data=obj, dtype=base.guess_dtype(obj))
+                h5o.link(ds.id, self.id, name, lcpl=lcpl)
+
+        if do_link:
+            fn = filename_encode(obj.filename)
+            with phil:
+                self.id.links.create_external(name, fn, self._e(obj.path),
+                                              lcpl=lcpl, lapl=self._lapl)
 
     @with_phil
     def __delitem__(self, name):
@@ -408,7 +516,7 @@ class Group(HLObject, MutableMappingWithLock):
 
         Returning None continues iteration, returning anything else stops
         and immediately returns that value from the visit method.  No
-        particular order of iteration within groups is guranteed.
+        particular order of iteration within groups is guaranteed.
 
         Example:
 
@@ -419,6 +527,7 @@ class Group(HLObject, MutableMappingWithLock):
         """
         with phil:
             def proxy(name):
+                """ Call the function with the text name, not bytes """
                 return func(self._d(name))
             return h5o.visit(self.id, proxy)
 
@@ -433,7 +542,7 @@ class Group(HLObject, MutableMappingWithLock):
 
         Returning None continues iteration, returning anything else stops
         and immediately returns that value from the visit method.  No
-        particular order of iteration within groups is guranteed.
+        particular order of iteration within groups is guaranteed.
 
         Example:
 
@@ -448,6 +557,7 @@ class Group(HLObject, MutableMappingWithLock):
         """
         with phil:
             def proxy(name):
+                """ Use the text name of the object, not bytes """
                 name = self._d(name)
                 return func(name, self[name])
             return h5o.visit(self.id, proxy)
@@ -455,16 +565,16 @@ class Group(HLObject, MutableMappingWithLock):
     @with_phil
     def __repr__(self):
         if not self:
-            r = six.u("<Closed HDF5 group>")
+            r = u"<Closed HDF5 group>"
         else:
             namestr = (
-                six.u('"%s"') % self.name
-            ) if self.name is not None else six.u("(anonymous)")
-            r = six.u('<HDF5 group %s (%d members)>') % (namestr, len(self))
+                u'"%s"' % self.name
+            ) if self.name is not None else u"(anonymous)"
+            r = u'<HDF5 group %s (%d members)>' % (namestr, len(self))
 
-        if six.PY3:
-            return r
-        return r.encode('utf8')
+        if six.PY2:
+            return r.encode('utf8')
+        return r
 
 
 class HardLink(object):
@@ -477,7 +587,6 @@ class HardLink(object):
     pass
 
 
-#TODO: implement equality testing for these
 class SoftLink(object):
 
     """
@@ -488,6 +597,7 @@ class SoftLink(object):
 
     @property
     def path(self):
+        """ Soft link value.  Not guaranteed to be a valid path. """
         return self._path
 
     def __init__(self, path):
@@ -506,15 +616,18 @@ class ExternalLink(object):
 
     @property
     def path(self):
+        """ Soft link path, i.e. the part inside the HDF5 file. """
         return self._path
 
     @property
     def filename(self):
+        """ Path to the external HDF5 file in the filesystem. """
         return self._filename
 
     def __init__(self, filename, path):
-        self._filename = str(filename)
-        self._path = str(path)
+        self._filename = filename_decode(filename_encode(filename))
+        self._path = path
 
     def __repr__(self):
-        return '<ExternalLink to "%s" in file "%s"' % (self.path, self.filename)
+        return '<ExternalLink to "%s" in file "%s"' % (self.path,
+                                                       self.filename)
